/*******************************************************************************
 * Copyright 2008, 2012 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package kdo.ebn.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Map;

import kdo.ebn.Competence;
import kdo.ebn.ExtendedBehaviorNetwork;
import kdo.ebn.IEBNAction;
import kdo.ebn.IEBNPerception;
import kdo.ebn.Proposition;
import kdo.ebn.xml.converter.CompetenceConverter;
import kdo.ebn.xml.converter.ExtendedBehaviorNetworkConverter;
import kdo.ebn.xml.converter.IBehaviorConverter;
import kdo.ebn.xml.converter.IBeliefConverter;
import kdo.ebn.xml.converter.PerceptionConverter;
import kdo.ebn.xml.converter.ResourceConverter;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.io.xml.DomDriver;

/**
 * Main class to serialize and deserialize networks to and from xml
 * @author Klaus Dorer
 */
public class NetworkSerializer
{
	private final XStream xStream;

	// TODO exception management??

	/**
	 * @param beliefs reference to the beliefs and resource beliefs
	 * @param behaviors reference to the behaviors
	 */
	public NetworkSerializer(Map<String, IEBNPerception> beliefs, Map<String, IEBNAction> behaviors)
	{
		xStream = new XStream(new DomDriver());
		xStream.registerConverter(new ExtendedBehaviorNetworkConverter(beliefs, behaviors));
		xStream.registerConverter(new IBeliefConverter(beliefs));
		xStream.registerConverter(new IBehaviorConverter(behaviors));
		xStream.registerConverter(new PerceptionConverter());
		xStream.registerConverter(new ResourceConverter());
		xStream.registerConverter(new CompetenceConverter());
		xStream.alias("Competence", Competence.class);
		xStream.alias("Proposition", Proposition.class);
		xStream.alias("EBN", ExtendedBehaviorNetwork.class);
	}

	/**
	 * Creates an ebn from the passed xml file
	 * @param filename the file containing an xml representation of an EBN
	 * @param beliefs the available beliefs to which to connect the perception
	 *        nodes
	 * @param behaviors the available behaviors to which to connect the actions
	 * @return a fully connected Extended Behavior network
	 */
	public static ExtendedBehaviorNetwork createEBN(
			String filename, Map<String, IEBNPerception> beliefs, Map<String, IEBNAction> behaviors)
	{
		NetworkSerializer serializer = new NetworkSerializer(beliefs, behaviors);
		return serializer.loadNetworkFromFile(new File(filename));
	}

	/**
	 * stores an ebn in a xml string
	 * @param ebn ebn to store
	 * @return xml-representation of the ebn
	 */
	public String storeNetwork(ExtendedBehaviorNetwork ebn)
	{
		return xStream.toXML(ebn);
	}

	/**
	 * stores an ebn in a file
	 * @param f target file
	 * @param ebn ebn to store
	 */
	public void storeNetworkToFile(File f, ExtendedBehaviorNetwork ebn)
	{
		String xml = storeNetwork(ebn);

		byte[] buffer = xml.getBytes();
		FileOutputStream fs;
		try {
			fs = new FileOutputStream(f);
			fs.write(buffer);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * loads an ebn from a xml string
	 * @param xml xml-representation of the ebn
	 * @return the ebn
	 */
	public ExtendedBehaviorNetwork loadNetwork(String xml)
	{
		return (ExtendedBehaviorNetwork) xStream.fromXML(xml);
	}

	/**
	 * loads an ebn from a file
	 * @param file xml-file
	 * @return the ebn
	 */
	public ExtendedBehaviorNetwork loadNetworkFromFile(File file)
	{
		byte[] buffer = new byte[(int) file.length()];
		FileInputStream fs;
		try {
			fs = new FileInputStream(file);
			fs.read(buffer);
			fs.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		String xml = new String(buffer);

		return loadNetwork(xml);
	}
}
