/*******************************************************************************
 * Copyright 2008, 2012 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package kdo.ebnDevKit.gui.ebnGraphPanel.guiElements.util;

import java.util.LinkedList;

/**
 * this class scales values to a range between 0 and 1. to reach this, it
 * calculates an internal factor. to reach this, it stores information about the
 * highest value in the last few rounds. a round means a usage cycle of this
 * scaler. the end of round must be marked by invoking markRound(). note that
 * all negative values return 0 as scaled value and will be used to calculate
 * the internal factor.
 * @author Thomas Rinklin
 */
public class ValueScaler
{
	private final int scaleRounds;

	private final LinkedList<Double> maxPerRound;

	/** current factor */
	private double factor = Double.MAX_VALUE;

	/**
	 * constructor: scaleRounds = 100
	 */
	public ValueScaler()
	{
		this(100);
	}

	/**
	 * constructor
	 * @param scaleRounds amount of rounds, which the object uses to calculation
	 *        the factor
	 */
	public ValueScaler(int scaleRounds)
	{
		this.scaleRounds = scaleRounds;
		maxPerRound = new LinkedList<Double>();
		maxPerRound.add(Double.MIN_VALUE);
	}

	/**
	 * scales a value to a range between 0 and 1 and use the value to calculate
	 * the internal factor.
	 * @param value the value to scale
	 * @return the scaled value
	 */
	public double scaleValue(double value)
	{
		return scaleValue(value, true);
	}

	/**
	 * scales a value to a range between 0 and 1
	 * @param value value to scale
	 * @param useToCalcFactor use this value to calculate the internal factor
	 * @return scaled value
	 */
	public double scaleValue(double value, boolean useToCalcFactor)
	{
		if (useToCalcFactor)
			updateRoundList(value);
		if (value < 0)
			return 0;
		if (value * factor > 1)
			return 1;
		return value * factor;
	}

	private void updateRoundList(double value)
	{
		if (value > maxPerRound.getFirst()) {
			maxPerRound.set(0, value);
		}
	}

	/**
	 * marks a round of scale usage
	 */
	public void markRound()
	{
		// calculate maximal value of the last (scaleRounds) rounds
		double maxVal = 0;
		for (Double maxValRound : maxPerRound) {
			maxVal = Math.max(maxVal, maxValRound);
		}

		// rescale
		// System.out.print(".");
		double scaledMaxVal = factor * maxVal;
		if (scaledMaxVal < 0.8 || scaledMaxVal > 1) {
			// System.out.print(factor);
			factor = 0.98 / maxVal;
			// System.out.println(" -> " + factor);
		}

		if (maxPerRound.size() >= scaleRounds)
			maxPerRound.removeLast();
		maxPerRound.addFirst(Double.MIN_VALUE);
	}
}
