package kdo.neuralNetwork;

import java.util.List;

import kdo.neuralNetwork.data.InOutPattern;

public interface INeuralNetwork {
	/**
	 * Initializes all weights of this network with values that are uniformly
	 * distributed between standardValue - delta (including) and standardValue +
	 * delta (not including)
	 * @param standardValue mean value for the weights
	 * @param delta maximal delta to the standardValue
	 */
	public abstract void initializeWeights(float standardValue, float delta);

	/**
	 * Calculates the output of this network for the input passed
	 * @param input the input for which to calculate the output. The length has
	 *        to fit to the number of inputs. The input has to include the bias
	 *        as 0th element
	 * @return
	 */
	public abstract float[] recall(float[] input);

	/**
	 * Recalls the passed input and returns the sum of square errors
	 * @param input the input for which to calculate the output. The length has
	 *        to fit to the number of inputs. The input has to include the bias
	 *        as 0th element
	 * @param output the expected output
	 * @return the sum of square errors
	 */
	public abstract float recallError(float[] input, float[] output);

	/**
	 * Recalls the full set of passed training data
	 * @param patterns the training patterns to recall
	 * @return the root of the mean squared errors
	 */
	double recallAllRMS(List<InOutPattern> patterns);

	/**
	 * Recalls the full set of passed training data and counts the number of
	 * missclassified patterns. Typically used with winner takes all output
	 * function.
	 * @param patterns the training patterns to recall
	 * @return the percentage of missclassified patterns
	 */
	double recallAllPercent(List<InOutPattern> patterns);

	/**
	 * Trains the passed pattern a single time
	 * @param input the input of the example
	 * @param expectedOutput the expected output against which to learn
	 * @return the error in this pattern as mean of all delta squares of output
	 *         and expectedOutput
	 */
	public abstract float train(float[] input, float[] expectedOutput);

	/**
	 * Trains the full set of passed training data sequentially
	 * @param patterns the training patterns to train
	 * @param maxLoops the number of loops to run through all patterns
	 * @return the root of the mean squared errors
	 */
	double trainAll(List<InOutPattern> patterns, int maxLoops);
}