/*
 * Copyright (c) 2008 Klaus Dorer
 * Hochschule Offenburg
 */
package kdo.neuralNetwork.feedforward;

import java.util.Arrays;

import kdo.neuralNetwork.output.IdentityOutputFunction;
import kdo.neuralNetwork.transfer.StepTransferFunction;
import kdo.util.IRandomSource;

/**
 * Represents a single layer Perceptron network
 */
public class Perceptron extends FeedForwardNetwork
{
	/**	 */
	private static final long serialVersionUID = 1L;

	/** the weights to the output layer */
	private final float[][] weight;

	/**
	 *
	 * @param noOfInputs the number of input neurons (including bias)
	 * @param noOfOutputs the number of output neurons
	 * @param initialLearnRate the initial value of the learnrate alpha
	 * @param rand the source for random numbers
	 */
	public Perceptron(int noOfInputs, int noOfOutputs, float initialLearnRate, IRandomSource rand)
	{
		super(initialLearnRate, rand, new StepTransferFunction(0.0f), new IdentityOutputFunction());
		if (noOfInputs < 1) {
			throw new IllegalArgumentException("Need at least one input but had: " + noOfInputs);
		}
		if (noOfOutputs < 1) {
			throw new IllegalArgumentException("Need at least one output but had: " + noOfOutputs);
		}
		weight = new float[noOfInputs][noOfOutputs];
		initializeWeights(0.0f, 1.0f);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see kdo.neuralNetwork.feedforward.INeuralNetwork#initializeWeights(float,
	 * float)
	 */
	public void initializeWeights(float standardValue, float delta)
	{
		for (float[] row : weight) {
			for (int j = 0; j < row.length; j++) {
				row[j] = random.nextFloat() * 2.0f * delta + standardValue - delta;
			}
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see kdo.neuralNetwork.feedforward.INeuralNetwork#recall(float[])
	 */
	public float[] recall(float[] input)
	{
		checkInputLength(input.length);

		int outputs = weight[0].length;
		float[] output = new float[outputs];
		for (int i = 0; i < outputs; i++) {
			for (int j = 0; j < weight.length; j++) {
				output[i] += weight[j][i] * input[j];
			}
			output[i] = transfer(output[i]);
		}
		return output;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see kdo.neuralNetwork.feedforward.INeuralNetwork#train(float[], float[])
	 */
	public float train(float[] input, float[] expectedOutput)
	{
		checkInputLength(input.length);
		checkOutputLength(expectedOutput.length);

		// calculate output
		float[] output = recall(input);

		// adjust weights
		float error = 0.0f;
		for (int i = 0; i < expectedOutput.length; i++) {
			for (int j = 0; j < weight.length; j++) {
				float delta = expectedOutput[i] - output[i];
				error += delta * delta;
				weight[j][i] += learnrate * input[j] * delta;
			}
		}

		System.out.println("Input: " + Arrays.toString(input) + " Output: " + Arrays.toString(output) +
						   " Expected: " + Arrays.toString(expectedOutput));
		error /= expectedOutput.length;
		return error;
	}

	/**
	 * @param inputLength the number of neurons in the input
	 */
	private void checkInputLength(int inputLength)
	{
		if (inputLength != weight.length) {
			throw new IllegalArgumentException(
					"Input size " + inputLength + " does not match network input size " + weight.length);
		}
	}

	/**
	 * @param outputLength the number of neurons in the output
	 */
	private void checkOutputLength(int outputLength)
	{
		if (outputLength != weight[0].length) {
			throw new IllegalArgumentException(
					"Output size " + outputLength + " does not match network output size " + weight[0].length);
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		return Arrays.deepToString(weight);
	}
}
