/**
 *
 */
package kdo.search;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import kdo.search.impl.SearchParameters;
import kdo.search.runs.averageOutStrategies.IAverageOutStrategy;
import kdo.search.runs.averageOutStrategies.IDomainVisitor;
import kdo.search.runs.averageOutStrategies.ParallelAverageOutStrategy;
import kdo.search.runs.averageOutStrategies.SerialAverageOutStrategy;
import kdo.search.runs.averageOutStrategies.ThreadedAverageOutStrategy;
import kdo.search.strategy.ILocalSearchStrategy;
import kdo.search.strategy.local.BasicVariableNeighborhoodSearch;
import kdo.search.strategy.local.FullSearch;
import kdo.search.strategy.local.HillClimbing;
import kdo.search.strategy.local.HillClimbingRandom;
import kdo.search.strategy.local.HillClimbingRandomRestart;
import kdo.search.strategy.local.RandomWalk;
import kdo.search.strategy.local.TabuSearch;
import kdo.search.strategy.local.UtilityMeasurement;
import kdo.search.strategy.local.annealing.LinearTemperatureReduction;
import kdo.search.strategy.local.annealing.SimulatedAnnealing;
import kdo.search.strategy.local.cmaes.impl.CMAESOptimizerAdapter;
import kdo.search.strategy.local.threaded.TabuSearchMultiThreaded;

/**
 * @author kdorer
 */
public class LearningConfigurationHelper
{
	private static final String MEASUREMENT = "Measurement";

	public static final String GENETIC = "Genetic";

	private static final String FULL_SEARCH = "FullSearch";

	private static final String CMAES_SEARCH = "CMAES";

	// local search strategies
	private static final String VARIABLE_NEIGHBORHOOD_SEARCH = "VNS";

	private static final String TABU_SEARCH = "TabuSearch";

	private static final String TABU_SEARCH_THREADED = "TabuSearchThreaded";

	private static final String SIMULATED_ANNEALING = "SimulatedAnnealing";

	private static final String RANDOM_WALK = "RandomWalk";

	private static final String HILL_CLIMBING_RANDOM_OPERATOR = "HCRandomOperator";

	private static final String HILL_CLIMBING_RANDOM_RESTART = "HCRandomRestart";

	private static final String HILL_CLIMBING = "HillClimbing";

	// avg out strategies
	public static final String SERIAL = "serial";

	private static final String PARALLEL = "parallel";

	private static final String THREADED = "threaded";

	/**
	 * A read-only list of all available optimization strategies.
	 */
	public static final List<String> STRATEGIES;

	static
	{
		List<String> strategies = new ArrayList<>();
		strategies.addAll(Arrays.asList(MEASUREMENT, GENETIC, FULL_SEARCH, CMAES_SEARCH, VARIABLE_NEIGHBORHOOD_SEARCH,
				TABU_SEARCH, TABU_SEARCH_THREADED, SIMULATED_ANNEALING, RANDOM_WALK, HILL_CLIMBING_RANDOM_OPERATOR,
				HILL_CLIMBING_RANDOM_RESTART, HILL_CLIMBING));
		STRATEGIES = Collections.unmodifiableList(strategies);
	}

	/**
	 * A read-only list of all available avg strategies.
	 */
	public static final List<String> AVG_STRATEGIES;

	static
	{
		List<String> avgStrategies = new ArrayList<>();
		avgStrategies.addAll(Arrays.asList(SERIAL, PARALLEL, THREADED));
		AVG_STRATEGIES = Collections.unmodifiableList(avgStrategies);
	}

	/**
	 *
	 * @param avgStrategy strategy name
	 * @param avgOutRuns number of runs
	 * @param threads the number of threads for parallel strategy
	 * @return an average out strategy for
	 */
	public static IAverageOutStrategy getAverageOutStrategy(
			String avgStrategy, int avgOutRuns, int threads, boolean verbose, IDomainVisitor visitor)
	{
		switch (avgStrategy) {
		case SERIAL:
			return new SerialAverageOutStrategy(avgOutRuns, verbose, visitor);
		case PARALLEL:
			return new ParallelAverageOutStrategy(avgOutRuns, threads, verbose, visitor);
		case THREADED:
			return new ThreadedAverageOutStrategy(avgOutRuns, verbose, visitor);
		}
		return null;
	}

	/**
	 * Returns a local search method to be used. Genetic search is treated
	 * separately, although in principle it is also a LocalSearch and properly
	 * inheriting.
	 * @param params local search parameters
	 * @param threads number of parallel threads
	 * @return the local search strategy to be used
	 */
	public static ILocalSearchStrategy getLocalSearchStrategy(SearchParameters params, int threads)
	{
		String strategyString = params.getStrategy();

		switch (strategyString) {
		case HILL_CLIMBING:
			return new HillClimbing(-1);
		case HILL_CLIMBING_RANDOM_RESTART:
			// do random operator selection
			return new HillClimbingRandomRestart(100, -1);
		case HILL_CLIMBING_RANDOM_OPERATOR:
			// do random operator selection
			return new HillClimbingRandom(1000000, 60000);
		case RANDOM_WALK:
			// do random walk
			return new RandomWalk(30, 100000, 60000);
		case SIMULATED_ANNEALING:
			return new SimulatedAnnealing(new LinearTemperatureReduction(0.5, 0.0), 1000, -1);
		case TABU_SEARCH:
			return new TabuSearch(100, 10000000, 60000000);
		case TABU_SEARCH_THREADED:
			return new TabuSearchMultiThreaded(100, 10000000, 60000000, threads);
		case VARIABLE_NEIGHBORHOOD_SEARCH:
			// do basic variable neighborhood search
			return new BasicVariableNeighborhoodSearch(new HillClimbing(-1), 175, 60000);
		case MEASUREMENT:
			return new UtilityMeasurement();
		case FULL_SEARCH:
			return new FullSearch();
		case CMAES_SEARCH:
			int populationSize = 50;
			int iterations = 70;
			double stopFitness = 10;
			double sigmaSpread = 0.1;
			return CMAESOptimizerAdapter.getInstance(populationSize, iterations, stopFitness, sigmaSpread);
		}
		return null;
	}
}
