/* Copyright 2009 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Rajit Shahi
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package kdo.search.runs.averageOutStrategies;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import kdo.domain.IIndividuum;
import kdo.domain.IUtilityCalculator;
import kdo.domain.UtilityCalculatorParameters;

/**
 * Base class for all average utility calculation strategies
 * @author kdorer
 */
public abstract class AverageOutStrategyBase implements IAverageOutStrategy
{
	/** the number of runs to perform to average utility on */
	protected final int averageOutRuns;

	protected final boolean verbose;

	private List<Map<String, Double>> propertiesHistory;

	/** the individual we work on */
	protected IIndividuum individuum;

	/** visitor that will be called to allow for domain specific implementation */
	protected IDomainVisitor visitor;

	/**
	 * @param averageOutRuns the number of runs to perform for the average
	 * @param verbose true if we print messages for each individual average out run
	 * @param visitor a visitor to inject domain specific implementations
	 */
	public AverageOutStrategyBase(int averageOutRuns, boolean verbose, IDomainVisitor visitor)
	{
		this.averageOutRuns = averageOutRuns;
		this.verbose = verbose;
		this.visitor = visitor;
	}

	protected IUtilityCalculator getUtilityCalculator(UtilityCalculatorParameters ucParameter)
	{
		return individuum.getUtilityCalculator(ucParameter);
	}

	protected void addValues(Map<String, Double> valuesToAdd, double utility, AverageOutState state)
	{
		if (verbose) {
			System.out.printf(Locale.US, "Run %d utility: %.2f\n", state.count, utility);
		}

		if (state.count == 0) {
			propertiesHistory = new ArrayList<>();
		}
		propertiesHistory.add(valuesToAdd);
		state.addValues(valuesToAdd, utility);
	}

	/**
	 * Called after all runs are finished
	 */
	protected void finishedCalculation(AverageOutState state)
	{
		System.out.printf(Locale.US, "Average utility: %5.3f averaged: %d, at %s, properties: ", state.averageUtility,
				state.count, new Date().toString());
		visitor.onFinishedCalculation(state.properties);
	}

	@Override
	public List<Map<String, Double>> getPropertiesHistory()
	{
		return propertiesHistory;
	}

	/**
	 * Override this to create domain specific parameters
	 * @param id the id for this run
	 * @return a utility calculation parameter instance
	 */
	protected UtilityCalculatorParameters createUCParameters(int id)
	{
		return visitor.createUCParameters(id);
	}
}
