/* Copyright 2009 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Rajit Shahi
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 */

package kdo.search.runs.averageOutStrategies;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import kdo.domain.IUtilityCalculator;
import kdo.domain.model.Individuum;

/**
 * Does a serial run of all trials, expecting that the server is already
 * running.
 * @author kdorer
 */
public class ParallelAverageOutStrategy extends AverageOutStrategyBase
{
	/** used to synchronize threads on */
	private Object synchroObject;

	/** the size of the thread pool */
	private int threadPoolSize;

	public ParallelAverageOutStrategy(int averageOutRuns, int threadPoolSize, boolean verbose, IDomainVisitor visitor)
	{
		super(averageOutRuns, verbose, visitor);
		this.threadPoolSize = threadPoolSize;
		synchroObject = new Object();
	}

	@Override
	public double getAverageUtility(Individuum individuum, AverageOutState state)
	{
		this.individuum = individuum;
		try {
			ExecutorService pool = Executors.newFixedThreadPool(threadPoolSize);

			Handler[] handler = new Handler[averageOutRuns];
			for (int i = 0; i < averageOutRuns; i++) {
				handler[i] = new Handler(i, state);
			}

			for (int i = 0; i < averageOutRuns; i++) {
				pool.execute(handler[i]);
			}

			pool.shutdown();
			pool.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
			synchronized (this)
			{
				// keep printing results synchronous
				handler[0].utilityCalculator.printParams();
				finishedCalculation(state);
			}

		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		return state.averageUtility;
	}

	/**
	 * Creates and checks for one neighborhood state if it is better than others
	 */
	class Handler implements Runnable
	{
		private int id;

		IUtilityCalculator utilityCalculator;

		private AverageOutState state;

		Handler(int id, AverageOutState state)
		{
			this.id = id;
			this.state = state;
		}

		@Override
		public void run()
		{
			double util = 0.0f;
			int repeatCount = 0;
			do {
				utilityCalculator = getUtilityCalculator(createUCParameters(id));
				util = utilityCalculator.singleRunUtility(repeatCount);
				repeatCount++;

				// we do not count invalid measurement runs
			} while (util == Double.NEGATIVE_INFINITY);

			// measures for this individuum
			synchronized (synchroObject)
			{
				addValues(utilityCalculator.getProperties(), util, state);
			}
		}
	}
}
